//
//  cocos2dAdventCalendarDay16World.m
//  cocos2dAdventCalendarDay16
//
//  Created by 岩井 雅幸 on 11/12/14.
//  Copyright myb design 2011年. All rights reserved.
//

#import "cocos2d.h"

extern "C" {
#import "CC3Foundation.h"
};

#import "cocos2dAdventCalendarDay16World.h"
#import "CC3PODResourceNode.h"
#import "CC3ActionInterval.h"
#import "CC3MeshNode.h"
#import "CC3Camera.h"
#import "CC3Light.h"
#import "CubePhysicsObject.h"


@implementation cocos2dAdventCalendarDay16World

/**
 * Constructs the 3D world.
 *
 * Adds 3D objects to the world, loading a 3D 'hello, world' message
 * from a POD file, and creating the camera and light programatically.
 *
 * When adapting this template to your application, remove all of the content
 * of this method, and add your own to construct your 3D model world.
 *
 * NOTE: The POD file used for the 'hello, world' message model is fairly large,
 * because converting a font to a mesh results in a LOT of triangles. When adapting
 * this template project for your own application, REMOVE the POD file 'hello-world.pod'
 * from the Resources folder of your project!!
 */
-(void) initializeWorld {

	// Create the camera, place it back a bit, and add it to the world
	CC3Camera* cam = [CC3Camera nodeWithName: @"Camera"];
	cam.location = cc3v( 0.0, 5.0, 12.0 );
	cam.rotation = cc3v(-35.0f, 0.0f, 0.0f);
	[self addChild: cam];

	// Create a light, place it back and to the left at a specific
	// position (not just directional lighting), and add it to the world
	CC3Light* lamp = [CC3Light nodeWithName: @"Lamp"];
	lamp.location = cc3v( -2.0, 0.0, 0.0 );
	lamp.isDirectionalOnly = NO;
	[cam addChild: lamp];

	// This is the simplest way to load a POD resource file and add the
	// nodes to the CC3World, if no customized resource subclass is needed.
//	[self addContentFromPODResourceFile: @"hello-world.pod"];
	[self addContentFromPODResourceFile: @"ccadcal16_cube.pod"];
	[self addContentFromPODResourceFile: @"ccadcal16_ground.pod"];
	
	// Create OpenGL ES buffers for the vertex arrays to keep things fast and efficient,
	// and to save memory, release the vertex data in main memory because it is now redundant.
	[self createGLBuffers];
	[self releaseRedundantData];
	
	// That's it! The model world is now constructed and is good to go.
	
	// If you encounter problems displaying your models, you can uncomment one or
	// more of the following lines to help you troubleshoot. You can also use these
	// features on a single node, or a structure of nodes. See the CC3Node notes.
	
	// Displays short descriptive text for each node (including class, node name & tag).
	// The text is displayed centered on the pivot point (origin) of the node.
//	self.shouldDrawAllDescriptors = YES;
	
	// Displays bounding boxes around those nodes with local content (eg- meshes).
//	self.shouldDrawAllLocalContentWireframeBoxes = YES;
	
	// Displays bounding boxes around all nodes. The bounding box for each node
	// will encompass its child nodes.
//	self.shouldDrawAllWireframeBoxes = YES;
	
	// Moves the camera so that it will display the entire scene.
//	[self.activeCamera moveWithDuration: 3.0 toShowAllOf: self];
	
	// If you encounter issues creating and adding nodes, or loading models from
	// files, the following line is used to log the full structure of the world.
	LogDebug(@"The structure of this world is: %@", [self structureDescription]);
	
	// ------------------------------------------

	// But to add some dynamism, we'll animate the 'hello, world' message
	// using a couple of cocos2d actions...
	
//	// Fetch the 'hello, world' 3D text object that was loaded from the
//	// POD file and start it rotating
//	CC3MeshNode* helloTxt = (CC3MeshNode*)[self getNodeNamed: @"Hello"];
//	CCActionInterval* partialRot = [CC3RotateBy actionWithDuration: 1.0
//														  rotateBy: cc3v(0.0, 30.0, 0.0)];
//	[helloTxt runAction: [CCRepeatForever actionWithAction: partialRot]];
//	
//	// To make things a bit more appealing, set up a repeating up/down cycle to
//	// change the color of the text from the original red to blue, and back again.
//	GLfloat tintTime = 8.0f;
//	ccColor3B startColor = helloTxt.color;
//	ccColor3B endColor = { 50, 0, 200 };
//	CCActionInterval* tintDown = [CCTintTo actionWithDuration: tintTime
//														  red: endColor.r
//														green: endColor.g
//														 blue: endColor.b];
//	CCActionInterval* tintUp = [CCTintTo actionWithDuration: tintTime
//														red: startColor.r
//													  green: startColor.g
//													   blue: startColor.b];
//	 CCActionInterval* tintCycle = [CCSequence actionOne: tintDown two: tintUp];
//	[helloTxt runAction: [CCRepeatForever actionWithAction: tintCycle]];
	
	// Prepare dynamics world
	collisionConfiguration = new btDefaultCollisionConfiguration();
	collisionDispatcher = new btCollisionDispatcher(collisionConfiguration);
	broadphaseInterface = new btDbvtBroadphase();
	constraintSolver = new btSequentialImpulseConstraintSolver();
	dynamicsWorld = new btDiscreteDynamicsWorld(collisionDispatcher, broadphaseInterface, constraintSolver, collisionConfiguration);
	dynamicsWorld->setGravity(btVector3(0.0f, -9.8f, 0.0f));
	
	// Create ground
	CC3MeshNode *groundMeshNode = (CC3MeshNode *)[self getNodeNamed:@"ground"];
	groundMeshNode.location = cc3v(0.0f, -5.0f, 0.0f);
	groundCollisionShape = new btBoxShape(btVector3(4.0f, 0.25f, 4.0f));
	btDefaultMotionState *groundMotionState = new btDefaultMotionState(btTransform(btQuaternion(0.0f, 0.0f, 0.0f, -1.0f), 
																				   btVector3(groundMeshNode.location.x, 
																							 groundMeshNode.location.y, 
																							 groundMeshNode.location.z)));
	btRigidBody::btRigidBodyConstructionInfo groundRigidBodyCI(0.0f, groundMotionState, groundCollisionShape, btVector3(0.0f, 0.0f, 0.0f));
	btRigidBody *groundRigidBody = new btRigidBody(groundRigidBodyCI);
	dynamicsWorld->addRigidBody(groundRigidBody);
	
	// Prepare cube template
	cubeMeshNode = (CC3MeshNode *)[self getNodeNamed:@"cube"];
	cubeMeshNode.visible = NO;
	cubeCollisionShape = new btBoxShape(btVector3(0.5f, 0.5f, 0.5f));
	cubes = [[NSMutableArray alloc] initWithCapacity:5];
	cubeGenerationTime = 0.0f;
	
	[[CCScheduler sharedScheduler] scheduleSelector:@selector(updatePhysics:) forTarget:self interval:0 paused:NO];
}


-(void) dealloc
{
	[cubes release];
	
	btCollisionObjectArray collisionObjects = dynamicsWorld->getCollisionObjectArray();
	for(int i=dynamicsWorld->getNumCollisionObjects()-1; i>=0; i--)
	{
		btCollisionObject* obj = collisionObjects[i];
		btRigidBody* rigidBody = btRigidBody::upcast(obj);
		if(rigidBody && rigidBody->getMotionState())
		{
			delete rigidBody->getMotionState();
		}
		dynamicsWorld->removeCollisionObject(obj);
		delete obj;
	}
	
	delete cubeCollisionShape;
	delete groundCollisionShape;
	
	delete dynamicsWorld;
	delete constraintSolver;
	delete broadphaseInterface;
	delete collisionDispatcher;
	delete collisionConfiguration;
	
	[super dealloc];
}


/**
 * This template method is invoked periodically whenever the 3D nodes are to be updated.
 *
 * This method provides this node with an opportunity to perform update activities before
 * any changes are applied to the transformMatrix of the node. The similar and complimentary
 * method updateAfterTransform: is automatically invoked after the transformMatrix has been
 * recalculated. If you need to make changes to the transform properties (location, rotation,
 * scale) of the node, or any child nodes, you should override this method to perform those
 * changes.
 *
 * The global transform properties of a node (globalLocation, globalRotation, globalScale)
 * will not have accurate values when this method is run, since they are only valid after
 * the transformMatrix has been updated. If you need to make use of the global properties
 * of a node (such as for collision detection), override the udpateAfterTransform: method
 * instead, and access those properties there.
 *
 * The specified visitor encapsulates the CC3World instance, to allow this node to interact
 * with other nodes in its world.
 *
 * The visitor also encapsulates the deltaTime, which is the interval, in seconds, since
 * the previous update. This value can be used to create realistic real-time motion that
 * is independent of specific frame or update rates. Depending on the setting of the
 * maxUpdateInterval property of the CC3World instance, the value of dt may be clamped to
 * an upper limit before being passed to this method. See the description of the CC3World
 * maxUpdateInterval property for more information about clamping the update interval.
 *
 * As described in the class documentation, in keeping with best practices, updating the
 * model state should be kept separate from frame rendering. Therefore, when overriding
 * this method in a subclass, do not perform any drawing or rending operations. This
 * method should perform model updates only.
 *
 * This method is invoked automatically at each scheduled update. Usually, the application
 * never needs to invoke this method directly.
 */
-(void) updateBeforeTransform: (CC3NodeUpdatingVisitor*) visitor {}

/**
 * This template method is invoked periodically whenever the 3D nodes are to be updated.
 *
 * This method provides this node with an opportunity to perform update activities after
 * the transformMatrix of the node has been recalculated. The similar and complimentary
 * method updateBeforeTransform: is automatically invoked before the transformMatrix
 * has been recalculated.
 *
 * The global transform properties of a node (globalLocation, globalRotation, globalScale)
 * will have accurate values when this method is run, since they are only valid after the
 * transformMatrix has been updated. If you need to make use of the global properties
 * of a node (such as for collision detection), override this method.
 *
 * Since the transformMatrix has already been updated when this method is invoked, if
 * you override this method and make any changes to the transform properties (location,
 * rotation, scale) of any node, you should invoke the updateTransformMatrices method of
 * that node, to have its transformMatrix, and those of its child nodes, recalculated.
 *
 * The specified visitor encapsulates the CC3World instance, to allow this node to interact
 * with other nodes in its world.
 *
 * The visitor also encapsulates the deltaTime, which is the interval, in seconds, since
 * the previous update. This value can be used to create realistic real-time motion that
 * is independent of specific frame or update rates. Depending on the setting of the
 * maxUpdateInterval property of the CC3World instance, the value of dt may be clamped to
 * an upper limit before being passed to this method. See the description of the CC3World
 * maxUpdateInterval property for more information about clamping the update interval.
 *
 * As described in the class documentation, in keeping with best practices, updating the
 * model state should be kept separate from frame rendering. Therefore, when overriding
 * this method in a subclass, do not perform any drawing or rending operations. This
 * method should perform model updates only.
 *
 * This method is invoked automatically at each scheduled update. Usually, the application
 * never needs to invoke this method directly.
 */
-(void) updateAfterTransform: (CC3NodeUpdatingVisitor*) visitor {}


- (void)generateCube
{
	GLfloat mass = 1.0f;
	btVector3 localInertia(0.0f, 0.0f, 0.0f);
	cubeCollisionShape->calculateLocalInertia(mass, localInertia);
	
	CC3MeshNode *meshNode = [cubeMeshNode copyAutoreleased];
	meshNode.location = cc3v(0.0f, 5.0f, 0.0f);
	meshNode.rotation = cc3v(rand() / (float)RAND_MAX * 90.0f, 
							 rand() / (float)RAND_MAX * 90.0f, 
							 rand() / (float)RAND_MAX * 90.0f);
	meshNode.visible = YES;
	[self addChild:meshNode];
	
	btDefaultMotionState *motionState = new btDefaultMotionState(btTransform(btQuaternion(meshNode.quaternion.x, 
																						  meshNode.quaternion.y, 
																						  meshNode.quaternion.z, 
																						  -meshNode.quaternion.w), 
																			 btVector3(meshNode.location.x, 
																					   meshNode.location.y, 
																					   meshNode.location.z)));
	btRigidBody::btRigidBodyConstructionInfo rigidBodyCI(mass, motionState, cubeCollisionShape, localInertia);
	btRigidBody *rigidBody = new btRigidBody(rigidBodyCI);
	dynamicsWorld->addRigidBody(rigidBody);
	
	CubePhysicsObject *cubeObj = [[CubePhysicsObject alloc] initWithMeshNode:meshNode rigidBody:rigidBody];
	[cubes addObject:cubeObj];
	[cubeObj release];
	
	[self updateWorld];
}


- (void)updatePhysics:(ccTime)dt
{
	dynamicsWorld->stepSimulation(dt, 2);
	
	NSEnumerator *enumerator = [cubes reverseObjectEnumerator];
	CubePhysicsObject *cube;
	while((cube = [enumerator nextObject]))
	{
		CC3MeshNode *meshNode = cube.meshNode;
		btRigidBody *rigidBody = cube.rigidBody;
		
		btTransform transform;
		rigidBody->getMotionState()->getWorldTransform(transform);
		btVector3 pos = transform.getOrigin();
		meshNode.location = cc3v(pos.getX(), pos.getY(), pos.getZ());
		btQuaternion quaternion = transform.getRotation();
		meshNode.quaternion = CC3Vector4Make(quaternion.getX(), quaternion.getY(), quaternion.getZ(), -quaternion.getW());
		
		if(meshNode.location.y < -30.0f)
		{
			// Remove cube from world
			[cubes removeObject:cube];
			[self removeChild:meshNode];
			delete rigidBody->getMotionState();
			dynamicsWorld->removeRigidBody(rigidBody);
			delete rigidBody;
		}
	}
	
	cubeGenerationTime += dt;
	if(cubeGenerationTime >= 2.0f)
	{
		[self generateCube];
		cubeGenerationTime = 0.0f;
	}
}

@end

